//
//  TftMDumbAI.java
//  UlimateRetro
//
//  Created by Billy Spelchan on Tue Oct 29 2002.
//  Copyright (c) 2002 Spelchan Software. All rights reserved.
//
package com.blazinggames.retro.modern.search;

// Java imports
import java.awt.Point;
import java.util.Vector;

// retro imports
import com.blazinggames.retro.shared.search.tftm.*;

/**
 * This is just the TftMAIManager class, but because the update wont happen
 * (at least on the mac) while ai code is executing I am doing the ai slightly
 * differently to have each move return.
*/
public class TftMDumbAI
{
// *********************
// ***** CONSTANTS *****
// *********************

    protected static final int MAX_TANKS = 9;


// *********************
// ***** VARIABLES *****
// *********************

    // configuration variables
    protected Map _map;
    protected int _id;
    protected int _money;
    protected int _turnMoney;
    protected int _factoryMoney;
    protected int _tankCost;

    // factory management
    protected Path _basicPaths[]; // Paths from base to factory
    protected int _factoryStates[]; // -1 = need, 0-8 = tank assigned, 10=owned
    protected Point _factoryPoints[];

    // Tank management
    protected Path _tankPaths[] = new Path[MAX_TANKS];

    // Game message management
    protected int _lastDay;
    protected String _message;

    // call management
    protected int _curTank;
    protected boolean _done;
    
// ************************
// ***** CONSTRUCTORs *****
// ************************

    public TftMDumbAI(Map map, int id, int money, int turnMoney, int factoryMoney, int tankCost)
    {
        _map = map;
        _id = id;
        _money = money;
        _turnMoney = turnMoney;
        _factoryMoney = factoryMoney;
        _tankCost = tankCost;
        initialize();
    }


// *****************************************
// ***** METHODS - Initialization aids *****
// *****************************************

    protected void initialize()
    {
        int numFactories = _map._factories.size();

        _basicPaths = new Path[numFactories];
        _factoryStates = new int[numFactories];
        _factoryPoints = new Point[numFactories];
        for (int cntr = 0; cntr < numFactories; ++cntr)
        {
            _factoryPoints[cntr] = new Point((Point)_map._factories.elementAt(cntr));
            _basicPaths[cntr] = new Path(_map, _map._bases[_id], _factoryPoints[cntr], 0);
            _factoryStates[cntr] = -1;
        }
    }


// ******************************
// ***** METHODS - Gameflow *****
// ******************************

    protected void updateGameState()
    {
        int cntr, temp;

        // remove any factorys that we lost
        for (cntr = 0; cntr < _factoryPoints.length; ++cntr)
        {
            temp = _map._map.getAttribute(_factoryPoints[cntr].x, _factoryPoints[cntr].y);
            if (_factoryStates[cntr] == 10)
            {
                if (temp != (MapTile.TILE_PLAYER1_FACTORY + _id))
                    _factoryStates[cntr] = -1;
            }
            else if (_factoryStates[cntr] >= 0)
            {
                if (_map._tanks[_id][_factoryStates[cntr]].getHealth() <= 0)
                    _factoryStates[cntr] = -1;
                if (temp == (MapTile.TILE_PLAYER1_FACTORY + _id))
                    _factoryStates[cntr] = 10;
            }
        }
    }

    /**
        * @return true if have enough money to build a tank.
     */
    protected boolean addMoney()
    {
        _money += _turnMoney;
        int temp = _factoryMoney * _map._playerFactoryCounts[_id];
        _money += temp;

        return (_tankCost < _money);
    }


    // ***************************************
    // ***** METHODS - Intelligence Aids *****
    // ***************************************

    protected void considerNewTank()
    {
        int cntr;

        if (_map.canCreateTank(_id))
        {
            Path target = null;
            int temp = 0;
            for (cntr = 0; cntr < _factoryPoints.length; ++cntr)
            {
                if (_factoryStates[cntr] == -1)
                {
                    if (target == null)
                    {
                        target = _basicPaths[cntr];
                        temp = cntr;
                    }
                    else if (target.getCost() > _basicPaths[cntr].getCost())
                    {
                        target = _basicPaths[cntr];
                        temp = cntr;
                    }
                }
            }

            if (target != null)
            {
                _factoryStates[temp] = _map.createTank(_id);
                _money -= _tankCost;
                _tankPaths[_factoryStates[temp]] = new Path(target);
                _message = "Created a new tank!";
            }
        }
    }

    protected boolean tankGuard(int tankNum, Point loc)
    {
        int los, targetDir = -1;

        for (int cntr = 0; cntr < 8; ++cntr)
        {
            los = _map.checkLOS(_id, tankNum, cntr);
            if ( (los >= 0) && (los != _id) )
                targetDir = cntr;
        }
        if (targetDir != -1)
        {
//            System.out.println("player " + _id + " tank " + tankNum + " found target...");
            int leftCost, rightCost, cost;
            int curFacing = _map._tanks[_id][tankNum].getFacing();
            leftCost = curFacing - targetDir;
            if (leftCost < 0)
                leftCost = 8 + curFacing - targetDir;
            rightCost = targetDir - curFacing;
            if (rightCost < 0)
                rightCost = 8 - curFacing + targetDir;
            cost = Math.min(leftCost, rightCost) * 10;
            if (cost < _map._tanks[_id][tankNum].getMovement())
            {
                _map._tanks[_id][tankNum].setFacing(targetDir);
                _map._tanks[_id][tankNum].useMovement(cost);
                //                _messages.addElement("[DEBUG] Turning from " + curFacing + " to " + targetDir + " costing " + cost);
            }
            else
                return false; // wait until next turn
            return tankShot(tankNum);
        }
        return false;
    }

    protected boolean tankTurn(int tankNum, Point loc, Point target)
    {
        int nface = 0, leftCost, rightCost, cost;
        int curFacing = _map._tanks[_id][tankNum].getFacing();

        if (loc.x < target.x)
        {
            if (loc.y < target.y)
            {
                nface = Tank.FACING_SOUTHEAST;
            }
            else if (loc.y == target.y)
            {
                nface = Tank.FACING_EAST;
            }
            else
            {
                nface = Tank.FACING_NORTHEAST;
            }
        }
        else if (loc.x == target.x)
        {
            if (loc.y < target.y)
            {
                nface = Tank.FACING_SOUTH;
            }
            else
            {
                nface = Tank.FACING_NORTH;
            }

        }
        else // loc.x > target.x
        {
            if (loc.y < target.y)
            {
                nface = Tank.FACING_SOUTHWEST;
            }
            else if (loc.y == target.y)
            {
                nface = Tank.FACING_WEST;
            }
            else
            {
                nface = Tank.FACING_NORTHWEST;
            }
        }
        leftCost = curFacing - nface;
        if (leftCost < 0)
            leftCost = 8 + curFacing - nface;
        rightCost = nface - curFacing;
        if (rightCost < 0)
            rightCost = 8 - curFacing + nface;
        cost = Math.min(leftCost, rightCost) * 10;
        if (cost < _map._tanks[_id][tankNum].getMovement())
        {
            _map._tanks[_id][tankNum].setFacing(nface);
            _map._tanks[_id][tankNum].useMovement(cost);
            //            _messages.addElement("[DEBUG] Turning from " + curFacing + " to " + nface + " costing " + cost);
        }
        else
        {
//            System.out.println("Not enough points to turn");
            return false; // wait until next turn
        }
        return true;
    }

    protected boolean tankMove(int tankNum, Point loc, Point target)
    {
        if (tankGuard(tankNum, loc) == true)
            return true;
        if (tankTurn(tankNum, loc, target) == false)
            return false;

        int moveLeft = _map._tanks[_id][tankNum].getMovement();
        int forwardCost = _map.costOfMoveForward(_id, tankNum);
        if (forwardCost <= moveLeft)
        {
            int moveResult = _map.moveTankForward(_id, tankNum);
            _message = "Moving Tank " + (tankNum + 1);
            if (moveResult > 0)
                _message = "Captured a factory!!!";
        }
        else
        {
//            System.out.println("Not enough movement to move forward");
            return false;            
        }
        return true;
    }


    protected boolean tankShot(int tankNum)
    {
//        System.out.println("Shooting at tank");
        int result = _map.fire(_id, tankNum);
        if (result < 1)
            return false;
        else if (result == 1)
        {
            _message = "Shot at enemy tank!";
        }
        else
        {
            _message = "Destroyed enemy tank!";
        }
        return true;
    }


    // ****************************************
    // ***** METHODS - Start the insanity *****
    // ****************************************

    public String startDay(int today)
    {
//        System.out.println("startDay() for " + _id + " called!");
        int cntr, sz;
        _message = "Analysing finances";
        _done = false;
        
        updateGameState();

        if (addMoney())
        {
            considerNewTank();
        }

        _curTank = -1;
        _done = false;
        nextTank();
        _lastDay = today;
        return _message;
    }

    protected String continueTurn()
    {
//        System.out.println("continueTurn() for " + _id + " called!");
        Point loc, target;
        boolean tankMoving;

        // Make sure there is actually a tank to continue moving
        if ( (_curTank < 0) || (_curTank > 8) )
        {
            nextTank();
            return " ";
        }
        else if (_map._tanks[_id][_curTank].getHealth() < 1)
        {
            nextTank();
            return " ";
        }
        
        _message = " ";
        loc = _map._tanks[_id][_curTank].getLocation();
        target = _tankPaths[_curTank].getNextNode(loc);
        if (loc.equals(target))
        {
            _message = "Guarding Location";
            tankMoving = tankGuard(_curTank, loc);
        }
        else
            tankMoving = tankMove(_curTank, loc, target);

        if (tankMoving == false)
            nextTank();
        return _message;
    }
    
    public boolean isDone()
    {
        return _done;
    }

    protected void nextTank()
    {
//        System.out.println("nextTank() for " + _id + " called! current tank " + _curTank);
        boolean foundNext = false;

        do
        {
            ++_curTank;
            if (_curTank > 8)
            {
                foundNext = true;
                _done = true;
            }
            else if (_map._tanks[_id][_curTank].getHealth() > 0)
            {
                foundNext = true;
                _map._tanks[_id][_curTank].startTurn();
//                System.out.println("Player " + _id + " tank " + _curTank + " movement is " + _map._tanks[_id][_curTank].getMovement());
            }
        }
        while (foundNext == false);
//        System.out.println("nextTank() for " + _id + " done at " + _curTank);
    }

    public int lastDay()
    {
        return _lastDay;
    }

    public int getCurrentTank()
    {
        return _curTank;
    }
}
