package com.blazinggames.retro.shared.search.tftm;

// java imports
import java.awt.Point;
import java.io.*;
import java.net.URL;
import java.util.Vector;

// spelchan imports
import com.blazinggames.urcl.ui.playfield.*;

/**
*/
public class Map 
{
// *********************
// ***** VARIABLES *****
// *********************

    // Map management
    public Playfield _map;
    public MapTile _tileRenderer;
    public int _mapWidth;
    public int _mapHeight;
    public Tank _tanks[][];

    // convienience variables
    public Point _bases[];
    public Vector _factories;
    public int _playerFactoryCounts[];


// ************************
// ***** CONSTRUCTORs *****
// ************************

    public Map()
    {
        prepareTileRenderer();
        prepareMap();
    }


// ********************************
// ***** METHODS - Setup aids *****
// ********************************

    /** 
     * This sets up the default tile renderer but is a separate method so
     * the renderer can be different in the modern version (if desired)
     * by simply overriding this method.
    */
    protected void prepareTileRenderer()
    {
        _tileRenderer = new MapTile();
    }
    
    /**
     * Prepares the map for use. While I don't think I'll need to change this
     * in the modern version, its nice to have the option.
    */
    protected void prepareMap()
    {
        int cntr, cntrTank;
        
        _map = new Playfield();
        _map.setTileset(5, 20, 20);
        for (cntr = 0; cntr < 5; ++cntr)
            _map.setTileRenderer(cntr, _tileRenderer);

        _tanks = new Tank[4][9];
    	_bases = new Point[4];
        _factories =  new Vector();
        _playerFactoryCounts = new int[4];
        for (cntr = 0; cntr < 4; ++cntr)
        {
            _bases[cntr] = new Point(0,0);
            for (cntrTank = 0; cntrTank < 9; ++cntrTank)
            {
                _tanks[cntr][cntrTank] = new Tank();
                _tanks[cntr][cntrTank].takeDamage(Integer.MAX_VALUE);
            }
        }
    }

    public void cloneMap(Map source)
    {
        int cntrRow, cntrCol, cell;
        
        prepareMap();
        _mapWidth = source._mapWidth;
        _mapHeight = source._mapHeight;
        _map.setPlayfield(_mapWidth, _mapHeight);

        for (cntrRow = 0; cntrRow < _mapHeight; ++cntrRow)
        {
            for (cntrCol = 0; cntrCol < _mapWidth; ++cntrCol)
            {
                cell =  source._map.getAttribute(cntrCol, cntrRow);
                _map.setTile(cntrCol, cntrRow, 0, cell);
                if (	(cell >= _tileRenderer.TILE_NEUTRAL_FACTORY) &&
                     (cell <= _tileRenderer.TILE_PLAYER4_FACTORY) )
                {
                    _factories.addElement(new Point(cntrCol, cntrRow));
                }
                else if (cell >= _tileRenderer.TILE_PLAYER1_BASE)
                {
                    _bases[cell - _tileRenderer.TILE_PLAYER1_BASE].setLocation(cntrCol, cntrRow);
                }
            }
        }
        
    }
    
    public void loadMap(URL url)
    throws IOException
    {
        int cntrRow, cntrCol, cell;
        String s;
        InputStream is = url.openStream();
        BufferedReader stream = new BufferedReader(new InputStreamReader(is));
        
        _mapWidth = Integer.parseInt(stream.readLine());
        _mapHeight = Integer.parseInt(stream.readLine());
        _map.setPlayfield(_mapWidth, _mapHeight);

        for (cntrRow = 0; cntrRow < _mapHeight; ++cntrRow)
        {
            s = stream.readLine();
            for (cntrCol = 0; cntrCol < _mapWidth; ++cntrCol)
            {
                cell =  _tileRenderer.parseSymbol(s.charAt(cntrCol));
                _map.setTile(cntrCol, cntrRow, 0, cell);
                if (	(cell >= _tileRenderer.TILE_NEUTRAL_FACTORY) &&
                        (cell <= _tileRenderer.TILE_PLAYER4_FACTORY) )
                {
                    _factories.addElement(new Point(cntrCol, cntrRow));
                }
                else if (cell >= _tileRenderer.TILE_PLAYER1_BASE)
                {
                    _bases[cell - _tileRenderer.TILE_PLAYER1_BASE].setLocation(cntrCol, cntrRow);
                }
            }
        }
        
        stream.close();
        
    }

// *****************************************
// ***** METHODS - getting and setting *****
// *****************************************

    public Tank getTankAt(int x, int y)
    {
        int id = _map.getTile(x, y) - 1;
        if (id < 0)
            return null;
        Tank r = null;

        for (int cntr = 0; cntr < 9; ++cntr)
        {
            if ( (_tanks[id][cntr].getX() == x) && (_tanks[id][cntr].getY() == y) )
            {
                r = _tanks[id][cntr];
                break;
            }
        }
        
        return r;
    }


    public boolean isTankOnBase(int id, int tanknum)
    {
        if ( (_bases[id].x == _tanks[id][tanknum].getX()) &&
             (_bases[id].y == _tanks[id][tanknum].getY()) )
            return true;
        else
            return false;
    }
    
// *************************************
// ***** METHODS - game management *****
// *************************************

    /** 
     * This is called after all the players haved moved.
     * It updates the factory counts and determins if there is a winner.
     *
     * @return player (0-3) who won, or -1 if no winner
    */
    public int endDay()
    {
        int cntr, temp;
        int numFactories = _factories.size();
        int winner = -1;
        Point p;
        
        // clear factory counts
        for (cntr = 0; cntr < 4; ++cntr)
            _playerFactoryCounts[cntr] = 0;
            
        // now go through factory counts
        for (cntr = 0; cntr < numFactories; ++cntr)
        {
            p = (Point)_factories.elementAt(cntr);
            temp = _map.getAttribute(p.x, p.y);
            if (temp > MapTile.TILE_NEUTRAL_FACTORY)
                ++_playerFactoryCounts[temp - MapTile.TILE_PLAYER1_FACTORY];
        }
        
        // now see if a player has all factories
        for (cntr = 0; cntr < 4; ++cntr)
            if (_playerFactoryCounts[cntr] == numFactories)
                winner = cntr;

        return winner;
    }

    public boolean canCreateTank(int id)
    {
        int cntr;
        boolean canCreate = false;
        
        // first see if a tank slot is empty
        for (cntr = 0; cntr < 9; ++cntr)
        {
            if (_tanks[id][cntr].getHealth() < 1)
            {
                canCreate = true;
                break;
            }
        }
        
        // now make sure there isn't a tank on the base
        if (_map.getTile(_bases[id].x, _bases[id].y) > 0)
            canCreate = false;

        return canCreate;
    }


    /**
     * Creates (activates) tank for player and places it on  the map.
     * The method assumes that a tank can be created (as this method should only
     * be called after a canCreateTank call).
     *
     * @param id the player (0-3) who is creating the tank
     * @return the slot (0-8) that the tank was assigned to
    */
    public int createTank(int id)
    {
        int tank = 0;
        
        // find lowest free tank
        for (int cntr = 8; cntr > -1; --cntr)
        {
            if (_tanks[id][cntr].getHealth() < 1)
                tank = cntr;
        }
        _tanks[id][tank].restock();
        _tanks[id][tank].setLocation(_bases[id].x, _bases[id].y);
        _tanks[id][tank].setFacing(0);
        _map.setTile(_bases[id].x, _bases[id].y, id + 1, _map.getAttribute(_bases[id].x, _bases[id].y));
        
        return tank;
    }

    public int costOfMoveForward(int id, int tankID)
    {
        Tank tank = _tanks[id][tankID];
        int oldX = tank.getX();
        int oldY = tank.getY();
        Point p = Tank.getMove(oldX, oldY, tank.getFacing());
        int t = _map.getTile(p.x, p.y);
        if (t > 0)
            return Integer.MAX_VALUE;
        int a = _map.getAttribute(p.x, p.y);
        int tspeed = _tileRenderer.getTravelSpeed(a);
        if ( 	(tspeed == 0) ||
              ( (a >= MapTile.TILE_PLAYER1_BASE) && ((a - MapTile.TILE_PLAYER1_BASE) != id)) )
            return Integer.MAX_VALUE;
        return tspeed;  
    }
    
    public int moveTankForward(int id, int tankID)
    {
        Tank tank = _tanks[id][tankID];
        int oldX = tank.getX();
        int oldY = tank.getY();
        Point p = Tank.getMove(oldX, oldY, tank.getFacing());
        if ( (p.x < 0) || (p.y < 0) || (p.x >= _mapWidth) || (p.y >= _mapHeight) )
            return -1;
        int t = _map.getTile(p.x, p.y);
        if (t > 0)
            return -1;
        int a = _map.getAttribute(p.x, p.y);
        if ( 	(_tileRenderer.getTravelSpeed(a) == 0) ||
                ( (a >= MapTile.TILE_PLAYER1_BASE) && ((a - MapTile.TILE_PLAYER1_BASE) != id)) )
            return -1;
        _map.setTile(oldX, oldY, 0, _map.getAttribute(oldX,oldY));
        tank.moveForeward(_tileRenderer.getTravelSpeed(a));
        int r = 0;
        if ( (a >= MapTile.TILE_NEUTRAL_FACTORY) &&
             (a <= MapTile.TILE_PLAYER4_FACTORY) &&
             (a != (MapTile.TILE_PLAYER1_FACTORY + id)) )
        {
            a = MapTile.TILE_PLAYER1_FACTORY + id;
            r = 1;
        }
        _map.setTile(p.x, p.y, id + 1, a);
        
        if (a >= MapTile.TILE_PLAYER1_BASE)
            tank.restock();
            
        return r;
    }

    public int moveTankBackward(int id, int tankID)
    {
        return -1;
    }

    /**
     * Launches a projectile and follows its path, damaging or destroying
     * a tank if it hits one. Returns one of the following codes:
     * <UL>
     * <LI> -1 no ammo so can't fire
     * <LI> 0 missed
     * <LI> 1 hit tank
     * <LI> 2 destroyed tank
     * </UL>
     *
     * @param id player who is firing
     * @param tankID specific tank that is firing
     * @return one of the above result codes
    */
    public int fire(int id, int tankID)
    {
        Tank tank = _tanks[id][tankID];
        if (tank.fire() ==  false)
            return -1;
        Point p = new Point(tank.getX(), tank.getY());
        int facing = tank.getFacing();
        int power = tank.getPower();
        int result = 0;
        do
        {
            p = Tank.getMove(p.x, p.y, facing);
            if ( (p.x < 0) || (p.y < 0) || (p.x >= _mapWidth) || (p.y >= _mapHeight) )
            {
                power = 0;
            }
            else if (_map.getTile(p.x, p.y) > 0)
            {
                Tank hitTank = getTankAt(p.x, p.y);
                hitTank.takeDamage(1);
                if (hitTank.getHealth() > 0)
                {
                    result = 1;
                    power = 0;
                }
                else
                {
                    result = 2;
                    _map.setTile(p.x, p.y, 0, _map.getAttribute(p.x, p.y));
                    power = 0;
                }
            }
            else
            {
                power -= _tileRenderer.getBulletCost(_map.getAttribute(p.x, p.y));
            }
        }
        while (power > 0);
        
        return result;
    }


    /**
     * Follows a potential projectiles path to see if there is a tank in range.
     * Returns one of the following codes:
     * <UL>
     * <LI> -1 no ammo so can't fire or no tank in los
     * <LI> 0..4 id of tank that would be hit
     * </UL>
     *
     * @param id player who is firing
     * @param tankID specific tank that is firing
     * @param dir LOS direction to check.
     * @return one of the above result codes
     */
    public int checkLOS(int id, int tankID, int dir)
    {
        Tank tank = _tanks[id][tankID];
        if (tank.getAmmo() <= 0)
            return -1;
        Point p = new Point(tank.getX(), tank.getY());
        int power = tank.getPower();
        int result = -1;
        do
        {
            p = Tank.getMove(p.x, p.y, dir);
            if ( (p.x < 0) || (p.y < 0) || (p.x >= _mapWidth) || (p.y >= _mapHeight) )
            {
                power = 0;
            }
            else if (_map.getTile(p.x, p.y) > 0)
            {
                    result = _map.getTile(p.x, p.y) - 1;
                    power = 0;
            }
            else
            {
                power -= _tileRenderer.getBulletCost(_map.getAttribute(p.x, p.y));
            }
        }
        while (power > 0);

        return result;
    }

    
// **********************************************
// ***** METODS - text version game Display *****
// **********************************************

    public String renderRow(int row)
    {
        StringBuffer sb = new StringBuffer();
        
        for (int cntr = 0; cntr < _mapWidth; ++cntr)
        {
            if (_map.getTile(cntr, row) == 0)
                sb.append(_tileRenderer.getSymbol(_map.getAttribute(cntr, row)));
            else
                sb.append(_map.getTile(cntr, row));
        }
        
        return sb.toString();
    }


// *****************************
// ***** METHODS - Testing *****
// *****************************

    private void test()
    {
        _map.setPlayfield(20,20);
        for (int cntrRow = 0; cntrRow < 16; ++cntrRow)
            for (int cntrCol = 0; cntrCol < 20; ++cntrCol)
                _map.setTile(cntrRow, cntrCol, 0, cntrRow);
        _mapWidth = 20;
        _mapHeight = 20;
    }
}
