package com.blazinggames.retro.shared.search.tftm;

// java imports
import java.awt.Point;

/**
*/
public class Tank 
{
// *********************
// ***** CONSTANTS *****
// *********************

    // Direction constants
    public static final int FACING_NORTH = 0;
    public static final int FACING_NORTHEAST = 1;
    public static final int FACING_EAST = 2;
    public static final int FACING_SOUTHEAST = 3;
    public static final int FACING_SOUTH = 4;
    public static final int FACING_SOUTHWEST = 5;
    public static final int FACING_WEST = 6;
    public static final int FACING_NORTHWEST = 7;


// **********************
// ***** VARIABLES ******
// **********************

    private int _maxHealth;
    private int _maxAmmo;
    private int _movementPerTurn;
    private int _power;
    private int _turnCost;
    private int _fireCost;

    private int _x;
    private int _y;
    private int _facing;    
    private int _movement;
    private int _health;
    private int _ammo;


// ************************
// ***** CONSTRUCTORs *****
// ************************

    public Tank()
    {
        this(2, 10, 50, 5, 10, 30);
    }

    public Tank(int health, int ammo, int mpt, int power, int turn, int fire)
    {
        setTankType(health, ammo, mpt, power, turn, fire);
    }


// ******************************************************
// ***** METHODS - Information getting and Settings *****
// ******************************************************

    public void setTankType(int health, int ammo, int mpt, int power, int turn, int fire)
    {
        _maxHealth = health;
    	_maxAmmo = ammo;
    	_movementPerTurn = mpt;
    	_power = power;
        _turnCost = turn;
        _fireCost = fire;
        restock();
    }

    public int getPower()
    {
        return _power;
    }

    public int getFireCost()
    {
        return _fireCost;
    }
    
    public void setLocation(int x, int y)
    {
        _x = x;
        _y = y;
    }
    
    public int getX()
    {
        return _x;
    }
    
    public int getY()
    {
        return _y;
    }

    public Point getLocation()
    {
        return (new Point(_x, _y));
    }
    
    public void setFacing(int facing)
    {
        _facing = facing;
    }
    
    public int getFacing()
    {
        return _facing;
    }

    public int getMovement()
    {
        return _movement;
    }

    public int getHealth()
    {
        return _health;
    }

    public int getAmmo()
    {
        return _ammo;
    }


// *****************************
// ***** METHODS - Actions *****
// *****************************

    public void restock()
    {
        _health = _maxHealth;
    	_ammo = _maxAmmo;
        _movement = 0;
    }
    
    /**
     * Turns tank counter clockwise one direction (45 degrees).
     *
     * @return new facing (see constants)
    */
    public int turnLeft()
    {
        --_facing;
        _facing &= 7;
        _movement -= _turnCost;
        return _facing;
    }
    
    /**
     * Turns tank clockwise one direction (45 degrees).
     *
     * @return new facing (see constants)
    */
    public int turnRight()
    {
        ++_facing;
        _facing &= 7;
        _movement -= _turnCost;
        return _facing;
    }
    
    public void moveForeward(int terrainCost)
    {
        switch (_facing)
        {
            case FACING_NORTH:
                --_y;
                break;
                
            case FACING_NORTHEAST:
                --_y;
                ++_x;
                break;
                
            case FACING_EAST:
                ++_x;
                break;
                
            case FACING_SOUTHEAST:
                ++_y;
                ++_x;
                break;
                
            case FACING_SOUTH:
                ++_y;
                break;
                
            case FACING_SOUTHWEST:
                ++_y;
                --_x;
                break;
                
            case FACING_WEST:
                --_x;
                break;
                
            case FACING_NORTHWEST:
                --_x;
                --_y;
                break;
        }
        _movement -= terrainCost;
    }
    
    public void moveBackwards(int terrainCost)
    {
        int oldFacing = _facing;
        _facing = (_facing + 4) & 7;
        moveForeward(terrainCost * 2);
        _facing = oldFacing;
    }
    
    /**
     * Note, this just lets you know if firing is possible and does the 
     * accounting (deducting bullets and movement) if it is, but the
     * actual bullet movement must be handled by the caller.
     *
     * @return true if successfully fired bullet, false otherwise.
    */
    public boolean fire()
    {
        if (_ammo > 0)
        {
            _movement -= _fireCost;
            --_ammo;
            return true;
        }
        else
            return false;
    }
    
    /**
     * called at the start of each turn. Note that overhead movement
     * (having a negative number of movement points) costs double the
     * amount and only half the unused points are retained (maximum
     * of full movement).
     *
     * @return number of movement points tank has
    */
    public int startTurn()
    {
        if (_movement < 0)
            _movement *= 2;
        else if (_movement > 1)
            _movement = Math.min(_movementPerTurn, _movement / 2);
            
        _movement += _movementPerTurn;
        if (_movement < 0)
            _movement = 0;
        
        return _movement;
    }
    
    /**
     * Removes indicated health from tank.
     *
     * @param n Amount of damage that the tank sustained
     * @return true if tank survived attack.
    */
    public boolean takeDamage(int n)
    {
        if (n >= _health)
        {
            _health = 0;
            _x = -1;
            _y = -1;
            return false;
        }
        _health -= n;
        return true;
    }

    public int useMovement(int n)
    {
        _movement -= n;
        return _movement;
    }
    
// *************************************************
// ***** METHODS - Testing and General Purpose *****
// *************************************************

    public static final Point getMove(int x, int y, int facing)
    {
        Point p = new Point(x, y);
        
        switch (facing)
        {
            case FACING_NORTH:
                --p.y;
                break;
                
            case FACING_NORTHEAST:
                --p.y;
                ++p.x;
                break;
                
            case FACING_EAST:
                ++p.x;
                break;
                
            case FACING_SOUTHEAST:
                ++p.y;
                ++p.x;
                break;
                
            case FACING_SOUTH:
                ++p.y;
                break;
                
            case FACING_SOUTHWEST:
                ++p.y;
                --p.x;
                break;
                
            case FACING_WEST:
                --p.x;
                break;
                
            case FACING_NORTHWEST:
                --p.x;
                --p.y;
                break;
        }

        return p;
    }

}
